#! /usr/bin/env python
############################################################################
#  classimgutils.py
#
#  Copyright 2016 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities which combine commands to create
#           useful extra functionality and make it more easily available
#           to be reused.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 17/12/2016
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

from __future__ import print_function

import numpy
import numpy.random

from osgeo import gdal

import rsgislib.imageutils
import rsgislib.imagecalc
import rsgislib.rastergis
import rsgislib
from rsgislib.classification import ClassSimpleInfoObj

import os
import os.path
import shutil
import random

from multiprocessing import Pool

from sklearn.model_selection import GridSearchCV
from sklearn.ensemble import RandomForestClassifier

import h5py

from rios import applier
from rios import cuiprogress
from rios import rat


def train_sklearn_classifer_gridsearch(classTrainInfo, paramSearchSampNum=0,
                                       gridSearch=GridSearchCV(RandomForestClassifier(), {})):
    """
A function to find the 'optimal' parameters for classification using a Grid Search
(http://scikit-learn.org/stable/modules/grid_search.html).
The returned classifier instance will be trained using all the inputted data.

:param classTrainInfo: list of rsgislib.classification.ClassSimpleInfoObj objects which will be used to
                       train the classifier.
:param paramSearchSampNum: the number of samples that will be randomly sampled from the training data for each class
                           for applying the grid search (tend to use a small data sample as can take a long time).
                           A value of 500 would use 500 samples per class.
:param gridSearch: is an instance of the sklearn.model_selection.GridSearchCV with an instance of the choosen
                   classifier and parameters to be searched.

    """
    if len(classTrainInfo) < 2:
        rsgislib.RSGISPyException(
            "Need at least 2 classes to be worth running findClassifierParametersAndTrain function.")

    first = True
    numVars = 0
    numVals = 0
    for classInfoVal in classTrainInfo.values():
        dataShp = h5py.File(classInfoVal.fileH5, 'r')['DATA/DATA'].shape
        if first:
            numVars = dataShp[1]
            first = False
        numVals += dataShp[0]

    dataArr = numpy.zeros([numVals, numVars], dtype=float)
    classArr = numpy.zeros([numVals], dtype=int)

    if paramSearchSampNum is not 0:
        dataArrSamp = numpy.zeros([(len(classTrainInfo) * paramSearchSampNum), numVars], dtype=float)
        classArrSamp = numpy.zeros([(len(classTrainInfo) * paramSearchSampNum)], dtype=int)

    rowInit = 0
    rowInitSamp = 0
    for key in classTrainInfo:
        # Open the dataset
        f = h5py.File(classTrainInfo[key].fileH5, 'r')
        numRows = f['DATA/DATA'].shape[0]
        # Copy data and populate classid array
        dataArr[rowInit:(rowInit + numRows)] = f['DATA/DATA']
        classArr[rowInit:(rowInit + numRows)] = classTrainInfo[key].id
        if paramSearchSampNum is not 0:
            # Create random index to sample the whole dataset.
            sampleIdxs = numpy.random.randint(0, high=numRows, size=paramSearchSampNum)
            # Sample the input data for classifier optimisation.
            dataArrSamp[rowInitSamp:(rowInitSamp + paramSearchSampNum)] = dataArr[rowInit:(rowInit + numRows)][
                sampleIdxs]
            classArrSamp[rowInitSamp:(rowInitSamp + paramSearchSampNum)] = classTrainInfo[key].id
            rowInitSamp += paramSearchSampNum
        rowInit += numRows
        f.close()

    if paramSearchSampNum is 0:
        dataArrSamp = dataArr
        classArrSamp = classArr

    print("Training data size: {} x {}".format(dataArr.shape[0], dataArr.shape[1]))
    print("Training Sample data size: {} x {}".format(dataArrSamp.shape[0], dataArrSamp.shape[1]))

    gridSearch.fit(dataArrSamp, classArrSamp)
    if not gridSearch.refit:
        raise Exception("Grid Search did no find a fit therefore failed...")

    print("Best score was {} and has parameters {}.".format(gridSearch.best_score_, gridSearch.best_params_))

    print('Training Classifier')
    gridSearch.best_estimator_.fit(dataArr, classArr)
    print("Completed")

    print('Calc Classifier Accuracy')
    accVal = gridSearch.best_estimator_.score(dataArr, classArr)
    print('Classifier Train Score = {}%'.format(round(accVal * 100, 2)))

    return gridSearch.best_estimator_


def train_sklearn_classifier(classTrainInfo, skClassifier):
    """
This function trains the classifier.

:param classTrainInfo: list of rsgislib.classification.ClassSimpleInfoObj objects which will be used to
                       train the classifier.
:param skClassifier: an instance of a parameterised scikit-learn classifier
                     (http://scikit-learn.org/stable/supervised_learning.html)

    """
    if len(classTrainInfo) < 2:
        rsgislib.RSGISPyException("Need at least 2 classes to be worth running trainClassifier function.")

    first = True
    numVars = 0
    numVals = 0
    for classInfoVal in classTrainInfo.values():
        dataShp = h5py.File(classInfoVal.fileH5, 'r')['DATA/DATA'].shape
        if first:
            numVars = dataShp[1]
            first = False
        numVals += dataShp[0]

    dataArr = numpy.zeros([numVals, numVars], dtype=float)
    classArr = numpy.zeros([numVals], dtype=int)

    rowInit = 0
    for key in classTrainInfo:
        # Open the dataset
        f = h5py.File(classTrainInfo[key].fileH5, 'r')
        numRows = f['DATA/DATA'].shape[0]
        # Copy data and populate classid array
        dataArr[rowInit:(rowInit + numRows)] = f['DATA/DATA']
        classArr[rowInit:(rowInit + numRows)] = classTrainInfo[key].id
        rowInit += numRows
        f.close()

    print("Training data size: {} x {}".format(dataArr.shape[0], dataArr.shape[1]))

    print('Training Classifier')
    skClassifier.fit(dataArr, classArr)
    print("Completed")

    print('Calc Classifier Accuracy')
    accVal = skClassifier.score(dataArr, classArr)
    print('Classifier Train Score = {}%'.format(round(accVal * 100, 2)))


def apply_sklearn_classifer(classTrainInfo, skClassifier, imgMask, imgMaskVal, imgFileInfo, outputImg, gdalformat,
                            classClrNames=True):
    """
This function uses a trained classifier and applies it to the provided input image.

:param classTrainInfo: dict (where the key is the class name) of rsgislib.classification.ClassSimpleInfoObj
                       objects which will be used to train the classifier (i.e., train_sklearn_classifier()),
                       provide pixel value id and RGB class values.
:param skClassifier: a trained instance of a scikit-learn classifier
                     (e.g., use train_sklearn_classifier or train_sklearn_classifer_gridsearch)
:param imgMask: is an image file providing a mask to specify where should be classified. Simplest mask is all
                the valid data regions (rsgislib.imageutils.genValidMask)
:param imgMaskVal: the pixel value within the imgMask to limit the region to which the classification is applied.
                   Can be used to create a heirachical classification.
:param imgFileInfo: a list of rsgislib.imageutils.ImageBandInfo objects (also used within
                    rsgislib.imageutils.extractZoneImageBandValues2HDF) to identify which images and bands are to
                    be used for the classification so it adheres to the training data.
:param outputImg: output image file with the classification. Note. by default a colour table and class names column
                  is added to the image. If an error is produced use HFA or KEA formats.
:param gdalformat: is the output image format - all GDAL supported formats are supported.
:param classClrNames: default is True and therefore a colour table will the colours specified in classTrainInfo
                      and a ClassName column (from imgFileInfo) will be added to the output file.

    """
    infiles = applier.FilenameAssociations()
    infiles.imageMask = imgMask
    numClassVars = 0
    for imgFile in imgFileInfo:
        infiles.__dict__[imgFile.name] = imgFile.fileName
        numClassVars = numClassVars + len(imgFile.bands)

    outfiles = applier.FilenameAssociations()
    outfiles.outimage = outputImg
    otherargs = applier.OtherInputs()
    otherargs.classifier = skClassifier
    otherargs.mskVal = imgMaskVal
    otherargs.numClassVars = numClassVars
    otherargs.imgFileInfo = imgFileInfo

    try:
        import tqdm
        progress_bar = rsgislib.TQDMProgressBar()
    except:
        progress_bar = cuiprogress.GDALProgressBar()

    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False

    # RIOS function to apply classifer
    def _applySKClassifier(info, inputs, outputs, otherargs):
        """
        Internal function for rios applier. Used within applyClassifer.
        """
        outClassVals = numpy.zeros_like(inputs.imageMask, dtype=numpy.uint32)
        if numpy.any(inputs.imageMask == otherargs.mskVal):
            outClassVals = outClassVals.flatten()
            imgMaskVals = inputs.imageMask.flatten()
            classVars = numpy.zeros((outClassVals.shape[0], otherargs.numClassVars), dtype=numpy.float)
            # Array index which can be used to populate the output array following masking etc.
            ID = numpy.arange(imgMaskVals.shape[0])
            classVarsIdx = 0
            for imgFile in otherargs.imgFileInfo:
                imgArr = inputs.__dict__[imgFile.name]
                for band in imgFile.bands:
                    classVars[..., classVarsIdx] = imgArr[(band - 1)].flatten()
                    classVarsIdx = classVarsIdx + 1
            classVars = classVars[imgMaskVals == otherargs.mskVal]
            ID = ID[imgMaskVals == otherargs.mskVal]
            predClass = otherargs.classifier.predict(classVars)
            outClassVals[ID] = predClass
            outClassVals = numpy.expand_dims(
                outClassVals.reshape((inputs.imageMask.shape[1], inputs.imageMask.shape[2])), axis=0)
        outputs.outimage = outClassVals

    print("Applying the Classifier")
    applier.apply(_applySKClassifier, infiles, outfiles, otherargs, controls=aControls)
    print("Completed")
    rsgislib.rastergis.populateStats(clumps=outputImg, addclrtab=True, calcpyramids=True, ignorezero=True)

    if classClrNames:
        ratDataset = gdal.Open(outputImg, gdal.GA_Update)
        red = rat.readColumn(ratDataset, 'Red')
        green = rat.readColumn(ratDataset, 'Green')
        blue = rat.readColumn(ratDataset, 'Blue')
        ClassName = numpy.empty_like(red, dtype=numpy.dtype('a255'))

        for classKey in classTrainInfo:
            print("Apply Colour to class \'" + classKey + "\'")
            red[classTrainInfo[classKey].id] = classTrainInfo[classKey].red
            green[classTrainInfo[classKey].id] = classTrainInfo[classKey].green
            blue[classTrainInfo[classKey].id] = classTrainInfo[classKey].blue
            ClassName[classTrainInfo[classKey].id] = classKey

        rat.writeColumn(ratDataset, "Red", red)
        rat.writeColumn(ratDataset, "Green", green)
        rat.writeColumn(ratDataset, "Blue", blue)
        rat.writeColumn(ratDataset, "ClassName", ClassName)
        ratDataset = None


def perform_voting_classification(skClassifiers, trainSamplesInfo, imgFileInfo, classAreaMask, classMaskPxlVal, tmpDIR,
                                  tmpImgBase, outClassImg, gdalformat='KEA', numCores=-1):
    """
A function which will perform a number of classification creating a combined classification by a simple vote.
The classifier parameters can be differed as a list of classifiers is provided (the length of the list is equal
to the number of votes), where the training data is resampled for each classifier. The analysis can be performed
using multiple processing cores.

Where:

:param skClassifiers: a list of classifiers (from scikit-learn), the number of classifiers defined
                      will be equal to the number of votes.
:param trainSamplesInfo: a list of rsgislib.classification.classimgutils.SamplesInfoObj objects used to
                         parameters the classifer and extract training data.
:param imgFileInfo: a list of rsgislib.imageutils.ImageBandInfo objects (also used within
                    rsgislib.imageutils.extractZoneImageBandValues2HDF) to identify which images and bands are
                    to be used for the classification so it adheres to the training data.
:param classAreaMask: a mask image which is used to specified the areas of the scene which are to be classified.
:param classMaskPxlVal: is the pixel value within the classAreaMask image for the areas of the image
                        which are to be classified.
:param tmpDIR: a temporary file location which will be created and removed during processing.
:param tmpImgBase: the same name of files written to the tmpDIR
:param outClassImg: the final output image file.
:param gdalformat: the output file format for outClassImg
:param numCores: is the number of processing cores to be used for the analysis (if -1 then all cores on the machine will be used).

Example::

    classVoteTemp = os.path.join(imgTmp, 'ClassVoteTemp')

    imgFileInfo = [rsgislib.imageutils.ImageBandInfo(img2010dB, 'sardb', [1,2]), rsgislib.imageutils.ImageBandInfo(imgSRTM, 'srtm', [1])]
    trainSamplesInfo = []
    trainSamplesInfo.append(SamplesInfoObj(className='Water', classID=1, maskImg=classTrainRegionsMask, maskPxlVal=1, outSampImgFile='WaterSamples.kea', numSamps=500, samplesH5File='WaterSamples_pxlvals.h5', red=0, green=0, blue=255))
    trainSamplesInfo.append(SamplesInfoObj(className='Land', classID=2, maskImg=classTrainRegionsMask, maskPxlVal=2, outSampImgFile='LandSamples.kea', numSamps=500, samplesH5File='LandSamples_pxlvals.h5', red=150, green=150, blue=150))
    trainSamplesInfo.append(SamplesInfoObj(className='Mangroves', classID=3, maskImg=classTrainRegionsMask, maskPxlVal=3, outSampImgFile='MangroveSamples.kea', numSamps=500, samplesH5File='MangroveSamples_pxlvals.h5', red=0, green=153, blue=0))

    skClassifiers = []
    for i in range(5):
        skClassifiers.append(ExtraTreesClassifier(n_estimators=50))

    for i in range(5):
        skClassifiers.append(ExtraTreesClassifier(n_estimators=100))

    for i in range(5):
        skClassifiers.append(ExtraTreesClassifier(n_estimators=50, max_depth=2))

    for i in range(5):
        skClassifiers.append(ExtraTreesClassifier(n_estimators=100, max_depth=2))

    mangroveRegionClassImg = MangroveRegionClass.kea
    classsklearn.perform_voting_classification(skClassifiers, trainSamplesInfo, imgFileInfo, classWithinMask, 1, classVoteTemp, 'ClassImgSample', mangroveRegionClassImg, gdalformat='KEA', numCores=-1)

    """

    def _apply_voting_classifier(inParams):
        """
        Internal function which is used by performVotingClassification
        """

        skClassifier = inParams['skClassifier']
        cTmpDIR = inParams['cTmpDIR']
        classAreaMask = inParams['classAreaMask']
        classMaskPxlVal = inParams['classMaskPxlVal']
        imgFileInfo = inParams['imgFileInfo']
        tmpClassImgOut = inParams['tmpClassImgOut']
        gdalformat = inParams['gdalformat']
        trainSamplesInfo = inParams['trainSamplesInfo']
        rndSeed = inParams['rndSeed']

        classTrainInfo = dict()
        for trainSamples in trainSamplesInfo:
            rsgislib.imageutils.performRandomPxlSampleInMaskLowPxlCount(inputImage=trainSamples.maskImg,
                                                                        outputImage=os.path.join(cTmpDIR,
                                                                        trainSamples.outSampImgFile),
                                                                        gdalformat=gdalformat,
                                                                        maskvals=[trainSamples.maskPxlVal],
                                                                        numSamples=trainSamples.numSamps,
                                                                        rndSeed=rndSeed)
            rsgislib.imageutils.extractZoneImageBandValues2HDF(imgFileInfo,
                                                               os.path.join(cTmpDIR, trainSamples.outSampImgFile),
                                                               os.path.join(cTmpDIR, trainSamples.samplesH5File),
                                                               trainSamples.maskPxlVal)
            classTrainInfo[trainSamples.className] = ClassSimpleInfoObj(id=trainSamples.classID,
                                                                        fileH5=os.path.join(cTmpDIR,
                                                                        trainSamples.samplesH5File),
                                                                        red=trainSamples.red,
                                                                        green=trainSamples.green,
                                                                        blue=trainSamples.blue)

        train_sklearn_classifier(classTrainInfo, skClassifier)
        apply_sklearn_classifer(classTrainInfo, skClassifier, classAreaMask, classMaskPxlVal, imgFileInfo,
                                tmpClassImgOut, gdalformat)

    rsgisUtils = rsgislib.RSGISPyUtils()

    if type(skClassifiers) is not list:
        raise rsgislib.RSGISPyException("A list of classifiers must be provided")

    numOfVotes = len(skClassifiers)

    if numCores <= 0:
        numCores = rsgisUtils.numProcessCores()

    tmpPresent = True
    if not os.path.exists(tmpDIR):
        os.makedirs(tmpDIR)
        tmpPresent = False

    outClassImgs = []
    mCoreParams = []
    dirs2DEL = []
    rndGen = random.seed()
    for i in range(numOfVotes):
        cTmpDIR = os.path.join(tmpDIR, str(i))
        if os.path.exists(cTmpDIR):
            shutil.rmtree(cTmpDIR, ignore_errors=True)
        os.makedirs(cTmpDIR)
        dirs2DEL.append(cTmpDIR)

        tmpClassImgOut = os.path.join(tmpDIR, tmpImgBase + '_' + str(i) + '.kea')
        outClassImgs.append(tmpClassImgOut)
        inParams = dict()
        inParams['skClassifier'] = skClassifiers[i]
        inParams['cTmpDIR'] = cTmpDIR
        inParams['classAreaMask'] = classAreaMask
        inParams['classMaskPxlVal'] = classMaskPxlVal
        inParams['imgFileInfo'] = imgFileInfo
        inParams['tmpClassImgOut'] = tmpClassImgOut
        inParams['gdalformat'] = 'KEA'
        inParams['trainSamplesInfo'] = trainSamplesInfo
        inParams['rndSeed'] = random.randrange(1000)
        mCoreParams.append(inParams)

    # Run processing on multiple cores.
    mProccesPool = Pool(numCores)
    mProccesPool.map(_apply_voting_classifier, mCoreParams)

    # Combine results using MODE.
    rsgislib.imagecalc.calcMultiImgBandStats(outClassImgs, outClassImg, rsgislib.SUMTYPE_MODE, gdalformat,
                                             rsgislib.TYPE_8UINT, 0, True)
    rsgislib.rastergis.populateStats(clumps=outClassImg, addclrtab=True, calcpyramids=True, ignorezero=True)

    # Colour output classification image.
    ratDataset = gdal.Open(outClassImg, gdal.GA_Update)
    red = rat.readColumn(ratDataset, 'Red')
    green = rat.readColumn(ratDataset, 'Green')
    blue = rat.readColumn(ratDataset, 'Blue')
    ClassName = numpy.empty_like(red, dtype=numpy.dtype('a255'))

    for trainSample in trainSamplesInfo:
        print("Apply Colour to class \'" + trainSample.className + "\'")
        red[trainSample.classID] = trainSample.red
        green[trainSample.classID] = trainSample.green
        blue[trainSample.classID] = trainSample.blue
        ClassName[trainSample.classID] = trainSample.className

    rat.writeColumn(ratDataset, "Red", red)
    rat.writeColumn(ratDataset, "Green", green)
    rat.writeColumn(ratDataset, "Blue", blue)
    rat.writeColumn(ratDataset, "ClassName", ClassName)
    ratDataset = None

    if not tmpPresent:
        shutil.rmtree(tmpDIR, ignore_errors=True)
    else:
        for cDIR in dirs2DEL:
            shutil.rmtree(cDIR, ignore_errors=True)







import rsgislib
import rsgislib.imageutils
import pprint
import math
import numpy
import osgeo.gdal as gdal


class StdImgBlockIter:

    """
    A class to read and write images in blocks where all blocks are all identical in size.
    To deal with image boundaries, where block do not divide into the the image size, the
    outside blocks will be overlapped with the previous block being defined with reference
    with the right and bottom edges rather than the left and top edges.
    """

    def __init__(self, img_info_lst, block_size):
        """

        :param img_info_lst:
        :param block_size:

        """
        self.img_info_lst = img_info_lst
        self.n_imgs = len(img_info_lst)
        self.block_size = block_size
        self.img_info = dict()
        self.define_overlaps()
        pprint.pprint(self.img_info)
        
        self.remain_x_block = False
        self.n_x_blocks = math.floor(self.overlap_width / block_size)
        if (self.n_x_blocks * block_size) < self.overlap_width:
            self.n_x_blocks += 1
            self.remain_x_block = True
        
        self.remain_y_block = False
        self.n_y_blocks = math.floor(self.overlap_height / block_size)
        if (self.n_y_blocks * block_size) < self.overlap_height:
            self.n_y_blocks += 1
            self.remain_y_block = True
        
        self.n_blocks = self.n_x_blocks * self.n_y_blocks
        
        self.c_blocks = 0
        self.c_x_blocks = -1
        self.c_y_blocks = 0

        for img in self.img_info_lst:
            self.img_info[img.name]['n_bands'] = len(img.bands)
            self.img_info[img.name]['np_arr'] = numpy.zeros((len(img.bands), block_size, block_size), dtype=float)
            self.img_info[img.name]['dataset'] = gdal.Open(img.fileName, gdal.GA_ReadOnly)
            if self.img_info[img.name]['dataset'] is None:
                raise Exception("Could not open image file: {}".format(img.fileName))
            self.img_info[img.name]['bands'] = dict()
            for band in img.bands:
                self.img_info[img.name]['bands'][band] = self.img_info[img.name]['dataset'].GetRasterBand(band)
                if self.img_info[img.name]['bands'][band] is None:
                    raise Exception("Could not open band {} within image {}".format(band, img.fileName))
        
        # Output image variables
        self.out_imgs_info = None

    def find_pxl_subset_bbox(self, spat_img_bbox, pxl_res, spat_sub_bbox):
        """

        :param spat_img_bbox:
        :param pxl_res:
        :param spat_sub_bbox:
        :return:

        """
        min_x_diff = spat_sub_bbox[0] - spat_img_bbox[0]
        max_y_diff = spat_img_bbox[3] - spat_sub_bbox[3]

        min_x_pxls = math.floor((min_x_diff / pxl_res[0]) + 0.5)
        max_y_pxls = math.floor((max_y_diff / pxl_res[1]) + 0.5)

        sub_width = math.floor(((spat_sub_bbox[1] - spat_sub_bbox[0]) / pxl_res[0]) + 0.5)
        sub_height = math.floor(((spat_sub_bbox[3] - spat_sub_bbox[2]) / pxl_res[1]) + 0.5)

        max_x_pxls = min_x_pxls + sub_width
        min_y_pxls = max_y_pxls + sub_height

        return [min_x_pxls, max_x_pxls, min_y_pxls, max_y_pxls]
    
    def reset_iter(self):
        """
        Reset the variables used for the iterator.

        """
        self.c_blocks = 0
        self.c_x_blocks = -1
        self.c_y_blocks = 0
    
    def define_overlaps(self):
        """
        An internal function used to calculate the amount of overlap between the input images.

        """
        rsgis_utils = rsgislib.RSGISPyUtils()
        bboxes = []
        self.img_epsg = None
        self.img_pxl_res = None
        for img in self.img_info_lst:
            self.img_info[img.name] = dict()
            self.img_info[img.name]['res'] = rsgis_utils.getImageRes(img.fileName)
            self.img_info[img.name]['bbox'] = rsgis_utils.getImageBBOX(img.fileName)
            self.img_info[img.name]['pxl_size'] = rsgis_utils.getImageSize(img.fileName)
            self.img_info[img.name]['epsg'] = rsgis_utils.getEPSGCode(img.fileName)
            if self.img_info[img.name]['epsg'] is None:
                raise Exception("The input image ({}) does not have a defined projection - please correct.".format(img.fileName))
            if self.img_epsg is None:
                self.img_epsg = self.img_info[img.name]['epsg']
            else:
                if self.img_epsg != self.img_info[img.name]['epsg']:
                    raise Exception("The input images have different input projections: EPSG:{} != EPSG:{}".format(self.img_epsg, self.img_info[img.name]['epsg']))
            
            if self.img_pxl_res is None:
                self.img_pxl_res = self.img_info[img.name]['res']
            else:
                if (self.img_pxl_res[0] != self.img_info[img.name]['res'][0]) and (self.img_pxl_res[1] != self.img_info[img.name]['res'][1]):
                    raise Exception("The input images have different input image resolutions: {} != {}".format(self.img_pxl_res, self.img_info[img.name]['res']))
            
            bboxes.append(self.img_info[img.name]['bbox'])
        
        self.bbox_intersect = rsgis_utils.bboxes_intersection(bboxes)
        
        self.overlap_width = 0
        self.overlap_height = 0
        
        for img in self.img_info_lst:
            self.img_info[img.name]['pxl_bbox'] = self.find_pxl_subset_bbox(self.img_info[img.name]['bbox'], self.img_info[img.name]['res'], self.img_info[img.name]['pxl_size'], self.bbox_intersect)
            if (self.overlap_width == 0) and (self.overlap_height == 0):
                self.overlap_width = self.img_info[img.name]['pxl_bbox'][1] - self.img_info[img.name]['pxl_bbox'][0]
                self.overlap_height = self.img_info[img.name]['pxl_bbox'][2] - self.img_info[img.name]['pxl_bbox'][3]
            else:
                overlap_width_tmp = self.img_info[img.name]['pxl_bbox'][1] - self.img_info[img.name]['pxl_bbox'][0]
                overlap_height_tmp = self.img_info[img.name]['pxl_bbox'][2] - self.img_info[img.name]['pxl_bbox'][3]
                if self.overlap_width != overlap_width_tmp:
                    raise Exception("The image widths do not match.")
                if self.overlap_height != overlap_height_tmp:
                    raise Exception("The image height do not match.")
            
    def __iter__(self):
        """
        Returns an instances of itself.
        """
        return self

    def __next__(self):
        """
        The key function for the iterator used to read the image blocks.

        :return: x_block, y_block, img_data (dict with image name as key; each image has
                 a np_arr variable for the image data.

        """
        if self.c_blocks < self.n_blocks:
            if self.c_x_blocks < (self.n_x_blocks-1):
                self.c_x_blocks += 1
            else:
                self.c_x_blocks = 0
                self.c_y_blocks += 1
            self.c_blocks += 1
            
            out_img_data = dict()
            
            if self.c_y_blocks == (self.n_y_blocks-1):
                if self.c_x_blocks == (self.n_x_blocks-1):
                    # X,Y End Case
                    for img in self.img_info_lst:
                        out_img_data[img.name] = dict()
                        out_img_data[img.name]['pxl_x'] = self.img_info[img.name]['pxl_bbox'][1] - self.block_size
                        out_img_data[img.name]['pxl_y'] = self.img_info[img.name]['pxl_bbox'][2] - self.block_size
                else:
                    # Y End Case
                    for img in self.img_info_lst:
                        out_img_data[img.name] = dict()
                        out_img_data[img.name]['pxl_x'] = self.img_info[img.name]['pxl_bbox'][0] + (self.block_size * self.c_x_blocks)
                        out_img_data[img.name]['pxl_y'] = self.img_info[img.name]['pxl_bbox'][2] - self.block_size
            elif self.c_x_blocks == (self.n_x_blocks-1):
                # X End Case.
                for img in self.img_info_lst:
                    out_img_data[img.name] = dict()
                    out_img_data[img.name]['pxl_x'] = self.img_info[img.name]['pxl_bbox'][1] - self.block_size
                    out_img_data[img.name]['pxl_y'] = self.img_info[img.name]['pxl_bbox'][3] + (self.block_size * self.c_y_blocks)
            else:
                # Not End Case
                for img in self.img_info_lst:
                    out_img_data[img.name] = dict()
                    out_img_data[img.name]['pxl_x'] = self.img_info[img.name]['pxl_bbox'][0] + (self.block_size * self.c_x_blocks)
                    out_img_data[img.name]['pxl_y'] = self.img_info[img.name]['pxl_bbox'][3] + (self.block_size * self.c_y_blocks)
                    
            for img in self.img_info_lst:
                for band in img.bands:
                    self.img_info[img.name]['bands'][band].ReadAsArray(xoff=out_img_data[img.name]['pxl_x'], yoff=out_img_data[img.name]['pxl_y'], win_xsize=self.block_size, win_ysize=self.block_size, buf_obj=self.img_info[img.name]['np_arr'][band-1])
                out_img_data[img.name]['np_arr'] = self.img_info[img.name]['np_arr']
                                            
            return (self.c_x_blocks, self.c_y_blocks, out_img_data)
        raise StopIteration
        
    def __len__(self):
        """
        A function which returns the length of the iterator (i.e., number of blocks).

        :return: number of blocks.

        """
        return self.n_blocks
    
    def create_output_imgs(self, out_imgs_info):
        """
        A function to create the output image(s) if required during processing.

        :param out_imgs_info: a list of rsgislib.imageutils.OutImageInfo objects specifying the output image(s).

        """
        rsgis_utils = rsgislib.RSGISPyUtils()
        out_wkt_str = rsgis_utils.getWKTFromEPSGCode(int(self.img_epsg))
        if out_wkt_str is None:
            raise Exception("Did not have a projection string for the EPSG code: {}".format(self.img_epsg))
        if math.fabs(self.img_pxl_res[0]) != math.fabs(self.img_pxl_res[1]):
            raise Exception("Currently limited to only outputting images with square pixels.")
        out_img_res = math.fabs(self.img_pxl_res[0])
        self.out_imgs_info = out_imgs_info
        self.out_imgs_objs = dict()
        for img in out_imgs_info:
            rsgislib.imageutils.createBlankImgFromBBOX(self.bbox_intersect, out_wkt_str, img.file_name, out_img_res, img.no_data_val, img.nbands, img.gdal_format, img.datatype, snap2grid=False)
            
            self.out_imgs_objs[img.name] = dict()
            self.out_imgs_objs[img.name]['dataset'] = gdal.Open(img.file_name, gdal.GA_Update)
            if self.out_imgs_objs[img.name]['dataset'] is None:
                raise Exception("Could not open image file: {}".format(img.file_name))
            self.out_imgs_objs[img.name]['nbands'] = img.nbands
            self.out_imgs_objs[img.name]['bands'] = dict()
            for band in range(img.nbands):
                self.out_imgs_objs[img.name]['bands'][band+1] = self.out_imgs_objs[img.name]['dataset'].GetRasterBand(band+1)
                if self.out_imgs_objs[img.name]['bands'][band+1] is None:
                    raise Exception("Could not open band {} within image {}".format(band+1, img.file_name))
            
    
    def write_block_to_image(self, x_block, y_block, img_blocks):
        """
        A function to write the processed blocks to an output image.

        :param x_block: The X index of the block being written.
        :param y_block: The Y index of the block being written.
        :param img_blocks: A dict of output image block data. Keys must match the name of the images.
                           Note. data block data returned must have shape (nbands, xsize, ysize)

        """
        if y_block == (self.n_y_blocks-1):
            if x_block == (self.n_x_blocks-1):
                # X,Y End Case
                pxl_x = self.overlap_width - self.block_size
                pxl_y = self.overlap_height - self.block_size
            else:
                # Y End Case
                pxl_x = self.block_size * x_block
                pxl_y = self.overlap_height - self.block_size
        elif x_block == (self.n_x_blocks-1):
            # X End Case.
            pxl_x = self.overlap_width - self.block_size
            pxl_y = self.block_size * y_block
        else:
            # Not End Case
            pxl_x = self.block_size * x_block
            pxl_y = self.block_size * y_block
        
        for img in img_blocks:
            if img in self.out_imgs_objs:
                block_shp = img_blocks[img].shape
                if block_shp[0] != self.out_imgs_objs[img]['nbands']:
                    raise Exception("The number of image bands in the output file and returned data block "
                                    "do not match (block:{}; image:{})".format(block_shp[0],
                                                                               self.out_imgs_objs[img]['nbands']))
                if (block_shp[1] != self.block_size) and (block_shp[2] != self.block_size):
                    raise Exception("The block size is either not square or the same size as the "
                                    "size expected (block: {} x {}; parameterised block: {}).".format(block_shp[1],
                                                                                                      block_shp[2],
                                                                                                      self.block_size))
                for band in range(block_shp[0]):
                    self.out_imgs_objs[img]['bands'][band+1].WriteArray(img_blocks[img][band], pxl_x, pxl_y)
                

"""
import tqdm

imgBandInfo = []
imgBandInfo.append(rsgislib.imageutils.ImageBandInfo(fileName='./S2_UVD_27sept_27700_sub.kea', name='sen2a', bands=[1,2,3,4,5,6,7,8,9,10]))
imgBandInfo.append(rsgislib.imageutils.ImageBandInfo(fileName='./S2_UVD_27sept_27700_sub_roi1.kea', name='sen2a1', bands=[1,2,3,4,5,6,7,8,9,10]))
imgBandInfo.append(rsgislib.imageutils.ImageBandInfo(fileName='./S2_UVD_27sept_27700_sub_roi2.kea', name='sen2a2', bands=[1,2,3,4,5,6,7,8,9,10]))

imgOutInfo = []
imgOutInfo.append(OutImageInfo(file_name='./S2_UVD_27sept_27700_sub_ndvi.kea', name='sen2a_ndvi', nbands=1, no_data_val=-999, gdal_format='KEA', datatype=rsgislib.TYPE_32FLOAT))
imgOutInfo.append(OutImageInfo(file_name='./S2_UVD_27sept_27700_sub_ndwi.kea', name='sen2a_ndwi', nbands=1, no_data_val=-999, gdal_format='KEA', datatype=rsgislib.TYPE_32FLOAT))

img_iter = StdImgBlockIter(imgBandInfo, 64)
img_iter.create_output_imgs(imgOutInfo)
for x_block, y_block, out_img_data in tqdm.tqdm(img_iter):
    hello = 'world'
    #print("Tile: ({} x {})".format(x_block, y_block))
    #for img in out_img_data:
    #    print("Mean = {}".format(numpy.mean(out_img_data[img]['np_arr'])))
    ndvi = (out_img_data['sen2a']['np_arr'][8] - out_img_data['sen2a']['np_arr'][3]) / (out_img_data['sen2a']['np_arr'][8] + out_img_data['sen2a']['np_arr'][3])
    ndwi = (out_img_data['sen2a']['np_arr'][8] - out_img_data['sen2a']['np_arr'][9]) / (out_img_data['sen2a']['np_arr'][8] + out_img_data['sen2a']['np_arr'][9])
    out_data = {'sen2a_ndvi': numpy.expand_dims(ndvi, axis=0), 'sen2a_ndwi': numpy.expand_dims(ndwi, axis=0)}
    #print(ndvi.shape)
    img_iter.write_block_to_image(x_block, y_block, out_data)

"""
